import numpy as np
from skimage.transform import radon, iradon

from .base import BaseTransform
from .cdt import CDT
from ..utils import check_array, assert_equal_shape
from ..utils import signal_to_pdf, match_shape2d


class RadonCDT(BaseTransform):
    """
    Radon Cumulative Distribution Transform.

    Parameters
    ----------
    theta : 1d array (default=np.arange(180))
        Radon transform projection angles.

    Attributes
    -----------
    displacements_ : array, shape (t, len(theta))
        Displacements u.
    transport_map_ : array, shape (t, len(theta))
        Transport map f.

    References
    ----------
    [The Radon cumulative distribution transform and its application to image
    classification]
    (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4871726/)
    """
    def __init__(self, theta=np.arange(180)):
        super(RadonCDT, self).__init__()
        self.theta = check_array(theta, ndim=1)
        self.epsilon = 1e-8
        self.total = 1.


    def forward(self, x0_range, sig0, x1_range, sig1, rm_edge=False):
        """
        Forward transform.

        Parameters
        ----------
        sig0 : array, shape (height, width)
            Reference image.
        sig1 : array, shape (height, width)
            Signal to transform.

        Returns
        -------
        rcdt : array, shape (t, len(theta))
            Radon-CDT of input image sig1.
        """
        # Check input arrays
        sig0 = check_array(sig0, ndim=2, dtype=[np.float64, np.float32],
                           force_strictly_positive=False)
        sig1 = check_array(sig1, ndim=2, dtype=[np.float64, np.float32],
                           force_strictly_positive=False)

        # Input signals must be the same size
        assert_equal_shape(sig0, sig1, ['sig0', 'sig1'])

        # Set reference signal
        self.sig0_ = sig0

        # Radon transform of signals
       
        rad1 = radon(sig1, theta=self.theta, circle=False)
        if len(np.unique(sig0)) == 1:
            rad0 = np.ones(rad1.shape)
        else:
            rad0 = radon(sig0, theta=self.theta, circle=False)

        # Initalize CDT, Radon-CDT, displacements, and transport map
        cdt = CDT()
        rcdt = []
        u = []
        f = []

        # Loop over angles
        for i in range(self.theta.size):
            # Convert projection to PDF
            rad0[0,i] = 0
            j0 = signal_to_pdf(rad0[:,i], epsilon=self.epsilon,
                               total=self.total)
            j1 = signal_to_pdf(rad1[:,i], epsilon=self.epsilon,
                               total=self.total)
            #x0=np.arange(len(j0))
            #x1=np.arange(len(j1))
            x0 = np.linspace(x0_range[0], x0_range[1], len(j0))
            x1 = np.linspace(x1_range[0], x1_range[1], len(j1))

            # Compute CDT of this projection
            lot, _,_ = cdt.forward(x0, j0, x1, j1, rm_edge)

            # Update 2D Radon-CDT, displacements, and transport map
            rcdt.append(lot)
            u.append(cdt.displacements_)
            f.append(cdt.transport_map_)

        # Convert lists to arrays
        rcdt = np.asarray(rcdt).T
        self.displacements_ = np.asarray(u).T
        self.transport_map_ = np.asarray(f).T

        self.is_fitted = True

        return rcdt


    def inverse(self, transport_map, sig0, x1_range):
        """
        Inverse transform.

        Returns
        -------
        sig1_recon : array, shape (height, width)
            Reconstructed signal sig1.
        """
        self._check_is_fitted()
        return self.apply_inverse_map(transport_map, sig0, x1_range)


    def apply_forward_map(self, transport_map, sig1):
        """
        Appy forward transport map.

        Parameters
        ----------
        transport_map : array, shape (t, len(theta))
            Forward transport map.
        sig1 : 2d array, shape (height, width)
            Signal to transform.

        Returns
        -------
        sig0_recon : array, shape (height, width)
            Reconstructed reference signal sig0.
        """
        # Check input arrays
        transport_map = check_array(transport_map, ndim=2,
                                    dtype=[np.float64, np.float32])
        sig1 = check_array(sig1, ndim=2, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)

        # Number of projections in transport map must match number of angles
        if transport_map.shape[1] != self.theta.size:
            raise ValueError("Length of theta must equal number of "
                             "projections in transport map: {} vs "
                             "{}".format(self.theta.size, transport_map.shape[1]))

        # Initialize Radon transforms
        rad1 = radon(sig1, theta=self.theta, circle=False)
        rad0 = np.zeros_like(rad1)

        # Check transport map and Radon transforms are the same size
        assert_equal_shape(transport_map, rad0,
                           ['transport_map', 'Radon transform of sig0'])

        # Loop over angles
        cdt = CDT()
        for i in range(self.theta.size):
            # Convert projection to PDF
            j1 = signal_to_pdf(rad1[:,i], epsilon=1e-8, total=1.)

            # Radon transform of sig0 comprised of inverse CDT of projections
            rad0[:,i] = cdt.apply_forward_map(transport_map[:,i], j1)

        # Inverse Radon transform
        sig0_recon = iradon(rad0, self.theta, circle=False, filter='ramp')

        # Crop sig0_recon to match sig1
        sig0_recon = match_shape2d(sig1, sig0_recon)

        return sig0_recon


    def apply_inverse_map(self, transport_map, sig0, x_range):
        """
        Appy inverse transport map.

        Parameters
        ----------
        transport_map : 2d array, shape (t, len(theta))
            Forward transport map. Inverse is computed in this function.
        sig0 : array, shape (height, width)
            Reference signal.

        Returns
        -------
        sig1_recon : array, shape (height, width)
            Reconstructed signal sig1.
        """
        # Check input arrays
        transport_map = check_array(transport_map, ndim=2,
                                    dtype=[np.float64, np.float32])
        sig0 = check_array(sig0, ndim=2, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)

        # Initialize Radon transforms
        if len(np.unique(sig0)) == 1:
            rad0 = np.ones(transport_map.shape)
        else:
            rad0 = radon(sig0, theta=self.theta, circle=False)
        rad1 = np.zeros_like(rad0)

        # Check transport map and Radon transforms are the same size
        assert_equal_shape(transport_map, rad0,
                           ['transport_map', 'Radon transform of sig0'])

        # Loop over angles
        cdt = CDT()
        for i in range(self.theta.size):
            # Convert projection to PDF
            j0 = signal_to_pdf(rad0[:,i], epsilon=1e-8, total=1.)
            
            x = np.linspace(x_range[0], x_range[1], len(j0))

            # Radon transform of sig1 comprised of inverse CDT of projections
            #rad1[:,i],_ = cdt.apply_inverse_map(transport_map[:,i], j0, x)
            rad1[:,i] = cdt.apply_inverse_map(transport_map[:,i], j0, x)

        # Inverse Radon transform
        sig1_recon = iradon(rad1, self.theta, circle=False, filter='ramp')

        # Crop sig1_recon to match sig0
        sig1_recon = match_shape2d(sig0, sig1_recon)

        return sig1_recon




import numpy as np
import numpy.linalg as LA
import multiprocessing as mp

from pytranskit.optrans.continuous.radoncdt import RadonCDT

eps = 1e-6
x0_range = [0, 1]
x_range = [0, 1]
Rdown = 4  # downsample radon projections (w.r.t. angles)
theta = np.linspace(0, 176, 180 // Rdown)

class RCDT_NS:
    def __init__(self, num_classes, thetas=theta, rm_edge=False):
        """
        Parameters
        ----------
        num_classes : integer, total number of classes
        thetas : array-like, angles in degrees for taking radon projections
            default = [0,180) with increment of 4 degrees.
        rm_edge : boolean flag; IF TRUE the first and last points of RCDTs will be removed
            default = False
        """
        self.num_classes = num_classes
        self.thetas = thetas
        self.rm_edge = rm_edge
        self.subspaces = []
        self.len_subspace = 0

    def fit(self, Xtrain, Ytrain, no_deform_model=False):
        """Fit linear model.
        Parameters
        ----------
        Xtrain : array-like, shape (n_samples, n_rows, n_columns)
            Image data for training.
        Ytrain : ndarray of shape (n_samples,)
            Labels of the training images.
        no_deform_model : boolean flag; IF TRUE, no deformation model will be added
            default = False.
        """
        
        # calculate the RCDT using parallel CPUs
        print('\nCalculating RCDTs for training images ...')
        Xrcdt = self.rcdt_parallel(Xtrain)
        
        # generate the basis vectors for each class
        print('Generating basis vectors for each class ...')
        for class_idx in range(self.num_classes):
            class_data = Xrcdt[Ytrain == class_idx]
            if no_deform_model:
                flat = class_data.reshape(class_data.shape[0], -1)
            else:
                class_data_trans = self.add_trans_samples(class_data)
                flat = class_data_trans.reshape(class_data_trans.shape[0], -1)
            
            u, s, vh = LA.svd(flat,full_matrices=False)
            
            cum_s = np.cumsum(s)
            cum_s = cum_s/np.max(cum_s)

            max_basis = (np.where(cum_s>=0.99)[0])[0] + 1
            
            if max_basis > self.len_subspace:
                self.len_subspace = max_basis
            
            basis = vh[:flat.shape[0]]
            self.subspaces.append(basis)


    def predict(self, Xtest, use_gpu=False):
        """Predict using the linear model
        Parameters
        ----------
        Xtest : array-like, shape (n_samples, n_rows, n_columns)
            Image data for testing.
        use_gpu: boolean flag; IF TRUE, use gpu for calculations
            default = False.
            
        Returns
        -------
        ndarray of shape (n_samples,)
           Predicted target values per element in Xtest.
        """
        
        # calculate the RCDT using parallel CPUs
        print('\nCalculating RCDTs for testing images ...')
        Xrcdt = self.rcdt_parallel(Xtest)
        
        # vectorize RCDT matrix
        X = Xrcdt.reshape([Xrcdt.shape[0], -1])
        
        # import cupy for using GPU
        if use_gpu:
            import cupy as cp
            X = cp.array(X)
        
        # find nearest subspace for each test sample
        print('Finding nearest subspace for each test sample ...')
        D = []
        for class_idx in range(self.num_classes):
            basis = self.subspaces[class_idx]
            basis = basis[:self.len_subspace,:]
            
            if use_gpu:
                D.append(cp.linalg.norm(cp.matmul(cp.matmul(X, cp.array(basis).T), 
                                                  cp.array(basis)) -X, axis=1))
            else:
                proj = X @ basis.T  # (n_samples, n_basis)
                projR = proj @ basis  # (n_samples, n_features)
                D.append(LA.norm(projR - X, axis=1))
        if use_gpu:
            preds = cp.argmin(cp.stack(D, axis=0), axis=0)
            return cp.asnumpy(preds)
        else:
            D = np.stack(D, axis=0)
            preds = np.argmin(D, axis=0)
            return preds


    def fun_rcdt_single(self, I):
        # I: (rows, columns)
        radoncdt = RadonCDT(self.thetas)
        template = np.ones(I.shape, dtype=I.dtype)
        Ircdt = radoncdt.forward(x0_range, template / np.sum(template), 
                                 x_range, I / np.sum(I), 
                                 self.rm_edge)
        return Ircdt
    
    def fun_rcdt_batch(self, data):
        # data: (n_samples, rows, columns)
        dataRCDT = [self.fun_rcdt_single(data[j, :, :] + eps) for j in range(data.shape[0])]
        return np.array(dataRCDT)
    
    def rcdt_parallel(self, X):
        # X: (n_samples, rows, columns)
        # calc RCDT of images
        n_cpu = np.min([mp.cpu_count(), X.shape[0]])
        splits = np.array_split(X, n_cpu, axis=0)
        pl = mp.Pool(n_cpu)
    
        dataRCDT = pl.map(self.fun_rcdt_batch, splits)
        rcdt_features = np.vstack(dataRCDT)  # (n_samples, proj_len, num_angles)
        pl.close()
        pl.join()

        return rcdt_features
        
    def add_trans_samples(self, rcdt_features):
        # rcdt_features: (n_samples, proj_len, num_angles)
        # deformation vectors for  translation
        v1, v2 = np.cos(self.thetas*np.pi/180), np.sin(self.thetas*np.pi/180)
        v1 = np.repeat(v1[np.newaxis], rcdt_features.shape[1], axis=0)
        v2 = np.repeat(v2[np.newaxis], rcdt_features.shape[1], axis=0)
        return np.concatenate([rcdt_features, v1[np.newaxis], v2[np.newaxis]])



import numpy as np
from scipy import interp
import matplotlib.pyplot as plt


from pytranskit.optrans.continuous.base import BaseTransform
from pytranskit.optrans.utils import check_array, assert_equal_shape, signal_to_pdf


class CDT(BaseTransform):
    """
    Cumulative Distribution Transform.

    Attributes
    -----------
    displacements_ : 1d array
        Displacements u.
    transport_map_ : 1d array
        Transport map f.

    References
    ----------
    [The cumulative distribution transform and linear pattern classification]
    (https://arxiv.org/abs/1507.05936)
    """
    def __init__(self):
        super(CDT, self).__init__()


    def forward(self, x0, sig0, x1, sig1, rm_edge=False):
        """
        Forward transform.

        Parameters
        ----------
        x0 : 1d array
            Independent axis variable of reference signal (sig0).
        sig0 : 1d array
            Reference signal.
        x1 : 1d array
            Independent axis variable of the signal to transform (sig1).
        sig1 : 1d array
            Signal to transform.

        Returns
        -------
        sig1_hat : 1d array
            CDT of input signal sig1.
        f : 1d array
            Forward transport map.
        xilde : 1d array
            Independent axis variable of sig1_hat.
        """
        # Check input arrays
        sig0 = check_array(sig0, ndim=1, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)
        sig1 = check_array(sig1, ndim=1, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)

        # Input signals must be the same size
        #assert_equal_shape(sig0, sig1, ['sig0', 'sig1'])

        self.sig0_ = sig0

        # Cumulative sums
        cum0 = np.cumsum(sig0)
        cum1 = np.cumsum(sig1)

        # x co-ordinates and interpolated y co-ordinates
        x = x1
        y = np.linspace(0, 1, sig0.size)
        y0 = interp(y, cum0, x0)    # inverse of CDF of sig0
        if len(np.unique(sig0)) == 1:
            y1 = interp(y, cum1, x)    # inverse of CDF of sig1 for uniform reference
        else:
            y1 = interp(cum0, cum1, x)    # inverse of CDF of sig1        

        # Compute displacements: u = f(x0)-x0
        self.displacements_ = interp(x0, y0, y1-y0)
        #self.displacements_ = y1 - x0

        # Compute transport map: f = u - x0
        #self.transport_map_ = self.displacements_ - x0
        self.transport_map_ = y1
        sig1_cdt = self.transport_map_
        self.xtilde = x0


        # CDT = (f - x) * sqrt(I0)
        sig1_hat = self.displacements_ * np.sqrt(sig0)
        
        if rm_edge:
            sig1_cdt = np.delete(sig1_cdt, 0)
            sig1_cdt = np.delete(sig1_cdt, len(sig1_cdt)-1)
            
            sig1_hat = np.delete(sig1_hat, 0)
            sig1_hat = np.delete(sig1_hat, len(sig1_hat)-1)
            
            y1 = np.delete(y1, 0)
            y1 = np.delete(y1, len(y1)-1)
            
            self.xtilde = np.delete(self.xtilde, 0)
            self.xtilde = np.delete(self.xtilde, len(self.xtilde)-1)
            
            self.transport_map_ = np.delete(self.transport_map_, 0)
            self.transport_map_ = np.delete(self.transport_map_, len(self.transport_map_)-1)

        self.is_fitted = True

        return sig1_cdt, sig1_hat, self.xtilde


    def inverse(self, transport_map, sig0, x1):
        """
        Inverse transform.
        
        Parameters
        ----------
        transport_map : 1d array
            Forward transport map.
        sig0 : 1d array
            Reference signal.
        x1 : 1d array
            Independent axis variable of the signal to reconstruct.

        Returns
        -------
        sig1_recon : 1d array
            Reconstructed signal.
        """
        self._check_is_fitted()
        return self.apply_inverse_map(transport_map, sig0, x1)


    def apply_forward_map(self, transport_map, sig1):
        """
        Appy forward transport map.

        Parameters
        ----------
        transport_map : 1d array
            Forward transport map.
        sig1 : 1d array
            Signal to transform.

        Returns
        -------
        sig0_recon : 1d array
            Reconstructed reference signal sig0.
        """
        # Check inputs
        transport_map = check_array(transport_map, ndim=1,
                                    dtype=[np.float64, np.float32])
        sig1 = check_array(sig1, ndim=1, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)
        assert_equal_shape(transport_map, sig1, ['transport_map', 'sig1'])

        # Reconstruct sig0
        x = np.arange(sig1.size)
        fprime = np.gradient(transport_map)
        sig0_recon = fprime * interp(transport_map, x, sig1)
        return sig0_recon


    def apply_inverse_map(self, transport_map, sig0, x):
        """
        Apply inverse transport map.

        Parameters
        ----------
        transport_map : 1d array
            Forward transport map. Inverse is computed in this function.
        sig0 : 1d array
            Reference signal.

        Returns
        -------
        sig1_recon : 1d array
            Reconstructed signal sig1.
        """
        # Check inputs
        transport_map = check_array(transport_map, ndim=1,
                                    dtype=[np.float64, np.float32])
        sig0 = check_array(sig0, ndim=1, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)
        assert_equal_shape(transport_map, sig0, ['transport_map', 'sig0'])

        # Reconstruct sig1
        
        fprime = np.gradient(transport_map)
        
        if len(np.unique(sig0)) == 1:
            sig1_recon = interp(x, transport_map, 1/fprime)
        else:
            sig1_recon = interp(x, transport_map, sig0/fprime)
            
        sig1_recon = signal_to_pdf(sig1_recon, epsilon=1e-7,
                               total=1.)

        return sig1_recon







import numpy as np
from scipy.io import loadmat
import os
import h5py
from PIL import Image
from sklearn.model_selection import train_test_split

def new_index_matrix(max_index, n_samples_perclass, num_classes, repeat, y_train):
    seed = int('{}{}{}'.format(n_samples_perclass, num_classes, repeat))
    np.random.seed(seed)
    index = np.zeros([num_classes, n_samples_perclass], dtype=np.int64)
    for classidx in range(num_classes):
        max_samples = (y_train == classidx).sum()
        index[classidx] = np.random.randint(0, max_samples, (n_samples_perclass))
    return index


def take_samples(data, labels, index, num_classes):
    assert data.shape[0] == labels.shape[0]
    assert index.shape[0] == num_classes
    indexed_data = []
    new_labels = []
    for i in range(num_classes):
       class_data, class_labels = data[labels == i], labels[labels == i]
       indexed_data.append(class_data[index[i]])
       new_labels.append(class_labels[index[i]])
    return np.concatenate(indexed_data), np.concatenate(new_labels)


def load_data(dataset, num_classes, datadir='data'):
    cache_file = os.path.join(datadir, dataset, 'dataset.hdf5')
    if os.path.exists(cache_file):
        with h5py.File(cache_file, 'r') as f:
            x_train, y_train = f['x_train'][()], f['y_train'][()]
            x_test, y_test = f['x_test'][()], f['y_test'][()]
            print('loaded from cache file data: x_train {} x_test {}'.format(x_train.shape, x_test.shape))
            return (x_train, y_train), (x_test, y_test)

    print('loading data from mat files')
    x_train, y_train, x_test, y_test = [], [], [], []
    for split in ['training', 'testing']:
        for classidx in range(num_classes):
            datafile = os.path.join(datadir, dataset, '{}/dataORG_{}.mat'.format(split, classidx))
            # loadmat(datafile)['xxO'] is of shape (H, W, N)
            data = loadmat(datafile)['xxO'].transpose([2, 0, 1]) # transpose to (N, H, W)
            label = np.zeros(data.shape[0], dtype=np.int64)+classidx
            #print('split {} class {} data.shape {}'.format(split, classidx, data.shape))
            if split == 'training':
                x_train.append(data)
                y_train.append(label)
            else:
                x_test.append(data)
                y_test.append(label)
    x_train, y_train = np.concatenate(x_train), np.concatenate(y_train)
    x_test, y_test = np.concatenate(x_test), np.concatenate(y_test)
    print('x_train.shape {} x_test.shape {}'.format(x_train.shape, x_test.shape))

    x_train = x_train / x_train.max(axis=(1, 2), keepdims=True)
    x_test = x_test / x_test.max(axis=(1, 2), keepdims=True)

    x_train = (x_train * 255.).astype(np.uint8)
    x_test = (x_test * 255.).astype(np.uint8)

    with h5py.File(cache_file, 'w') as f:
        f.create_dataset('x_train', data=x_train)
        f.create_dataset('y_train', data=y_train)
        f.create_dataset('x_test', data=x_test)
        f.create_dataset('y_test', data=y_test)
        print('saved to {}'.format(cache_file))

    return (x_train, y_train), (x_test, y_test)

def take_train_samples(x_train, y_train, n_samples_perclass, num_classes, repeat):
    max_index = x_train.shape[0] // num_classes
    train_index = new_index_matrix(max_index, n_samples_perclass, num_classes, repeat, y_train)
    x_train_sub, y_train_sub = take_samples(x_train, y_train, train_index, num_classes)
    return x_train_sub, y_train_sub

def take_train_val_samples(x_train, y_train, n_samples_perclass, num_classes, repeat):
    max_index = x_train.shape[0]//num_classes
    train_index = new_index_matrix(max_index, n_samples_perclass, num_classes, repeat, y_train)

    val_samples = n_samples_perclass // 10 # Use 10% for validation

    if val_samples >= 1:
        val_index = train_index[:, :val_samples]
        x_val, y_val = take_samples(x_train, y_train, val_index, num_classes)
        assert x_val.shape[0] == y_val.shape[0]
        print('validation data shape {}'.format(x_val.shape), end=' ')
    else:
        x_val, y_val = None, None
        print('validation data {}'.format(x_val), end=' ')

    train_sub_index = train_index[:, val_samples:]
    x_train_sub, y_train_sub = take_samples(x_train, y_train, train_sub_index, num_classes)
    print('train data shape {}'.format(x_train_sub.shape))

    if x_val is not None:
        assert x_val.shape[0] + x_train_sub.shape[0] == n_samples_perclass*num_classes
    else:
        assert x_train_sub.shape[0] == n_samples_perclass*num_classes


    return (x_train_sub, y_train_sub), (x_val, y_val)



def dataset_config(dataset):
    assert dataset in ['AffMNIST', 'LiverN', 'MNIST', 'OAM', 'OAM_t5', 'OAM_t10', 
                       'SignMNIST', 'Synthetic', 'CIFAR10', 'MNIST_outDist', 'HEP2']
    if dataset in ['MNIST']:
        rm_edge = True
        num_classes = 10
        po_train_max = 12  # maximum train samples = 2^po_max
        img_size = 28
    elif dataset in ['AffMNIST']:
        rm_edge = True
        num_classes = 10
        img_size = 84
        po_train_max = 12  # maximum train samples = 2^po_max
    elif dataset in ['OAM', 'OAM_t10', 'OAM_t5']:
        rm_edge = False
        num_classes = 32
        img_size = 151
        po_train_max = 9  # maximum train samples = 2^po_max
    elif dataset in ['SignMNIST']:
        rm_edge = False
        num_classes = 3
        img_size = 128
        po_train_max = 9  # maximum train samples = 2^po_max
    elif dataset in ['Synthetic']:
        rm_edge = True
        num_classes = 1000
        img_size = 64
        po_train_max = 4  # maximum train samples = 2^po_max
    elif dataset in ['LiverN']:
        rm_edge=False
        num_classes = 2
        img_size = 130
        po_train_max = 8   # maximum train samples = 2^po_max
    elif dataset in ['CIFAR10']:
        rm_edge = False
        num_classes = 10
        img_size = 32
        po_train_max = 12
    elif dataset in ['MNIST_outDist']:
        rm_edge = True
        num_classes = 10
        img_size = 84
        po_train_max = 12  # maximum train samples = 2^po_max
    elif dataset in ['HEP2']:
        rm_edge = False
        num_classes = 2
        img_size = 64
        po_train_max = 10  # maximum train samples = 2^po_max

    return num_classes, img_size, po_train_max, rm_edge



import six
import warnings
import numpy as np

def assert_all_finite(X):
    """
    Throw a ValueError if X contains NaN or infinity.
    """
    X = np.asanyarray(X)
    # First try an O(n) time, O(1) space solution for the common case that
    # everything is finite; fall back to O(n) space np.isfinite to prevent
    # false positives from overflow in sum method.
    if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
            and not np.isfinite(X).all()):
        raise ValueError("Input contains NaN, infinity"
                         " or a value too large for %r." % X.dtype)


def assert_equal_shape(a, b, names=None):
    """
    Throw a ValueError if a and b are not the same shape.
    """
    if names is None:
        names = ['a', 'b']

    if a.shape != b.shape:
        raise ValueError("{} and {} must be the same shape: "
                         "{!s} vs {!s}".format(names[0], names[1],
                         a.shape, b.shape))


def check_array(array, ndim=None, dtype='numeric', force_all_finite=True,
                force_strictly_positive=False):
    """
    Input validation on an array, list, or similar.

    Parameters
    ----------
    array : object
        Input object to check/convert
    ndim : int or None (default=None)
        Number of dimensions that array should have. If None, the dimensions
        are not checked
    dtype : string, type, list of types or None (default='numeric')
        Data type of result. If None, the dtype of the input is preserved.
        If 'numeric', dtype is preserved unless array.dtype is object.
        If dtype is a list of types, conversion on the first type is only
        performed if the dtype of the input is not in the list.
    force_all_finite : boolean (default=True)
        Whether to raise an error on np.inf and np.nan in array
    force_strictly_positive : boolean (default=False)
        Whether to raise an error if any array elements are <= 0

    Returns
    -------
    array_converted : object
        The converted and validated array.
    """
    # Do we want a numeric dtype?
    dtype_numeric = isinstance(dtype, six.string_types) and dtype == 'numeric'

    # Original dtype of the array
    dtype_orig = getattr(array, 'dtype', None)
    if not hasattr(dtype_orig, 'kind'):
        # Not a data type (e.g. a column named dtype in a pandas DataFrame)
        dtype_orig = None

    # If we want a numeric dtype
    if isinstance(dtype, six.string_types) and dtype == 'numeric':
        if dtype_orig is not None and dtype_orig.kind == "O":
            # If input is object, convert to float.
            dtype = np.float64
        else:
            dtype = None

    # If dtype is a list of possible types
    if isinstance(dtype, (list, tuple)):
        if dtype_orig is not None and dtype_orig in dtype:
            # No dtype conversion required
            dtype = None
        else:
            # dtype conversion required. Let's select the first element of the
            # list of accepted types.
            dtype = dtype[0]

    array = np.array(array, dtype=dtype)

    if ndim is not None and array.ndim != ndim:
        if ndim == 1 and array.ndim == 2 and array.shape[1] == 1:
            array = np.squeeze(array)
            warnings.warn("Converted column vector into 1D array.")
        else:
            raise ValueError("Expected {}D array, "
                             "got {}D array instead".format(ndim, array.ndim))

    # Check for NaNs and infs
    if force_all_finite:
        assert_all_finite(array)

    # Check that array is strictly positive
    if force_strictly_positive and np.any(array <= 0):
        raise ValueError("Array must be strictly positive (i.e. > 0).")

    return array


def check_decomposition(obj):
    """
    Check that an object is a PCA or PLDA (i.e. decomposition) object.

    Parameters
    ----------
    obj : object
        Object to check

    Returns
    -------
    mean : array, shape (n_features,)
        Mean of the data in the decomposition object
    components : array, shape (n_components, n_features)
        Components learned by the decomposition object
    std : array, shape (n_components,)
        Standard deviation of the training data projected on to each component
    """
    # Check that decomp is a PCA or PLDA object
    try:
        mean = obj.mean_
        components = obj.components_
        std = np.sqrt(obj.explained_variance_)
        return mean, components, std
    except AttributeError:
        raise



import numpy as np
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage import map_coordinates
from scipy.interpolate import griddata

from .validation import check_array, assert_equal_shape


def signal_to_pdf(input, sigma=0., epsilon=1e-8, total=1.):
    """
    Get the (smoothed) probability density function of a signal.

    Performs the following operations:
    1. Smooth sigma with a Gaussian filter
    2. Normalize signal such that it sums to 1
    3. Add epsilon to ensure signal is strictly positive
    4. Re-normalize signal such that it sums to total

    Parameters
    ----------
    input : ndarray
        Input array
    sigma : scalar
        Standard deviation for Gaussian kernel. The standard deviations of the
        Gaussian filter are given for each axis as a sequence, or as a single
        number, in which case it is equal for all axes.
    epsilon : scalar
        Offset to ensure that signal is strictly positive.
    total : scalar
        Value of the signal summation.

    Returns
    -------
    pdf : ndarray
        Returned array of same shape as input
    """
    input = check_array(input, dtype=['float32','float64'])

    if sigma < 0:
        raise ValueError('sigma must be >= 0.')

    if epsilon <= 0:
        raise ValueError('epsilon must be > 0.')

    if total <= 0:
        raise ValueError('total must be > 0.')

    pdf = gaussian_filter(input, sigma)
    pdf /= pdf.sum()
    pdf += epsilon
    pdf /= pdf.sum()
    pdf *= total
    return pdf


def match_shape2d(a, b):
    """
    Crop array B such that it matches the shape of A.

    Parameters
    ----------
    a : 2d array
        Array of desired size.
    b : 2d array
        Array to crop. Shape must be larger than (or equal to) the shape
        of array a.

    Returns
    -------
    b_crop : 2d array
        Cropped version of b, with the same shape as a.
    """
    a = check_array(a, ndim=2, force_all_finite=False)
    b = check_array(b, ndim=2, force_all_finite=False)

    # Difference in dimensions
    h, w = a.shape
    ylo = (b.shape[0] - h) // 2
    xlo = (b.shape[1] - w) // 2
    if ylo < 0 or xlo < 0:
        raise ValueError("A is bigger than B: "
                         "{!s} vs {!s}".format(a.shape, b.shape))

    return b[ylo:ylo+h,xlo:xlo+w]


def interp2d(img, f, order=1, fill_value=0.):
    """
    Interpolation for 2D gridded data.

    Parameters
    ----------
    img : 2d array, shape (height, width)
        Image to interpolate. This function assumes a grid of sample points:
        x = 0:(width-1), y = 0:(height-1).
    f : 3d array, shape (2, height, width)
        Coordinates of interpolated points. First dimension f[0] corresponds to
        y-coordinates, second dimension f[1] is x-coordinates.
    order : int (default=1)
        Order of the spline interpolation. Must be in the range 0-5.
    fill_value : float (default=0.)
        Value used for points outside the boundaries.

    Returns
    -------
    out, 2d array, shape (height, width)
        Image interpolated at points defined by f.
    """
    # Check inputs
    img = check_array(img, ndim=2, force_all_finite=True)
    f = check_array(f, ndim=3, force_all_finite=True)
    assert_equal_shape(img, f[0])

    # Interpolate
    h, w = img.shape
    out = map_coordinates(img, f.reshape((2,h*w)), order=order, cval=fill_value)

    return out.reshape((h,w))


def griddata2d(img, f, order=1, fill_value=0.):
    """
    Interpolate 2d scattered data

    Parameters
    ----------
    img : 2d array, shape (height, width)
        Image to interpolate.
    f : 3d array, shape (2, height, width)
        Coordinates of at which img is defined. First dimension f[0] corresponds
        to y-coordinates, second dimension f[1] is x-coordinates.
    order : int (default=1)
        Order of the interpolation. Must be in the range 0-2.
    fill_value : float (default=0.)
        Value used for points outside the boundaries.

    Returns
    -------
    out, 2d array, shape (height, width)
        Image interpolated on to regular gird defined by: x = 0:(width-1),
        y = 0:(height-1).
    """
    # Check inputs
    img = check_array(img, ndim=2, force_all_finite=True)
    f = check_array(f, ndim=3, force_all_finite=True)
    assert_equal_shape(img, f[0])

    # Determine interpolation type
    method = 'linear'
    if order == 0:
        method = 'nearest'
    elif order == 2:
        method = 'cubic'

    # Create grid of points
    h, w = img.shape
    xv, yv = np.meshgrid(np.arange(w,dtype=float), np.arange(h,dtype=float))

    # Interpolate image on to grid
    out = griddata(f.reshape((2,-1)).T, img.ravel(),
                   np.c_[yv.ravel(), xv.ravel()], method=method,
                   fill_value=fill_value)

    return out.reshape((h,w))



from .data_utils import signal_to_pdf, match_shape2d, interp2d, griddata2d
from .validation import check_array, assert_equal_shape, check_decomposition

__all__ = ['signal_to_pdf', 'match_shape2d', 'interp2d', 'griddata2d',
           'check_array', 'assert_equal_shape', 'check_decomposition']



import numpy as np

from ..utils import check_array, assert_equal_shape, interp2d, griddata2d


class BaseTransform(object):
    """
    Base class for optimal transport transform methods.

    .. warning::

       This class should **not** be used directly. Use derived classes instead.
    """
    def __init__(self):
        self.is_fitted = False
        self.sig0_ = None
        self.displacements_ = None
        self.transport_map_ = None


    def _check_is_fitted(self):
        if not self.is_fitted:
            raise AssertionError("The forward transform of {0!s} has not been "
                                 "called yet. Call 'forward' before using "
                                 "this method".format(type(self).__name__))


    def forward(self):
        """
        Placeholder for forward transform.
        Subclasses should implement this method!
        """
        raise NotImplementedError


    def inverse(self):
        """
        Inverse transform.

        Returns
        -------
        sig1_recon : array, shape (height, width)
            Reconstructed signal sig1.
        """
        self._check_is_fitted()
        return self.apply_inverse_map(self.transport_map_, self.sig0_)


    def apply_forward_map(self):
        """
        Placeholder for application of forward transport map.
        Subclasses should implement this method!
        """
        raise NotImplementedError


    def apply_inverse_map(self):
        """
        Placeholder for application of inverse transport map.
        Subclasses should implement this method!
        """
        raise NotImplementedError



class BaseMapper2D(BaseTransform):
    """
    Base class for 2D optimal transport transform methods (e.g. CLOT, VOT2D).

    .. warning::

       This class should **not** be used directly. Use derived classes instead.
    """
    def __init__(self):
        super(BaseMapper2D, self).__init__()
        return


    def apply_forward_map(self, transport_map, sig1):
        """
        Appy forward transport map.

        Parameters
        ----------
        transport_map : array, shape (2, height, width)
            Forward transport map.
        sig1 : array, shape (height, width)
            Signal to transform.

        Returns
        -------
        sig0_recon : array, shape (height, width)
            Reconstructed reference signal sig0.
        """
        # Check inputs
        transport_map = check_array(transport_map, ndim=3,
                                    dtype=[np.float64, np.float32])
        sig1 = check_array(sig1, ndim=2, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)
        assert_equal_shape(transport_map[0], sig1, ['transport_map', 'sig1'])

        # Jacobian and its determinant
        f0y, f0x = np.gradient(transport_map[0])
        f1y, f1x = np.gradient(transport_map[1])
        detJ = (f1x * f0y) - (f1y * f0x)

        # Reconstruct sig0
        sig0_recon = detJ * interp2d(sig1, transport_map, fill_value=sig1.min())

        return sig0_recon


    def apply_inverse_map(self, transport_map, sig0):
        """
        Appy inverse transport map.

        Parameters
        ----------
        transport_map : array, shape (2, height, width)
            Forward transport map. Inverse is computed in this function.
        sig0 : array, shape (height, width)
            Reference signal.

        Returns
        -------
        sig1_recon : array, shape (height, width)
            Reconstructed signal sig1.
        """
        # Check inputs
        transport_map = check_array(transport_map, ndim=3,
                                    dtype=[np.float64, np.float32])
        sig0 = check_array(sig0, ndim=2, dtype=[np.float64, np.float32],
                           force_strictly_positive=True)
        assert_equal_shape(transport_map[0], sig0, ['transport_map', 'sig0'])

        # Jacobian and its determinant
        f0y, f0x = np.gradient(transport_map[0])
        f1y, f1x = np.gradient(transport_map[1])
        detJ = (f1x * f0y) - (f1y * f0x)

        # Let's hope there are no NaNs/Infs in sig0/detJ
        sig1_recon = griddata2d(sig0/detJ, transport_map, fill_value=sig0.min())

        return sig1_recon



import setuptools

with open("README.md", "r") as fh:
    long_description=fh.read()

setuptools.setup(
  name="pytranskit",
  version="0.1.6",
  author="Abu Hasnat Mohammad Rubaiyat, Xuwang Yin, Liam Cattell",
  author_email="ar3fx@virginia.edu, xuwangyin@gmail.com",
  description="Python transport based signal processing toolkit.",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/rohdelab/PyTransKit",
  packages=setuptools.find_packages(),
  classifiers=[
    "Development Status :: 3 - Alpha",

    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",

    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
  ],
  python_requires='>=3.6',
)
